(    Title:  Dynamic-Strings word set tester
      File:  dstester.fs
 Test file:  dstester-test.fs
  Log file:  dstester.log
    Author:  David N. Williams
   License:  Public Domain, John Hayes
   Version:  1.4.0
   Revised:  February 11, 2018

For any code derived from John Hayes' tester program:
)
\ (C) 1995 JOHNS HOPKINS UNIVERSITY / APPLIED PHYSICS LABORATORY
\ MAY BE DISTRIBUTED FREELY AS LONG AS THIS COPYRIGHT NOTICE REMAINS.
(
The rest is public domain.

This code extends the John Hayes based ttester.fs to work with
our Dynamic-Strings word set, which has separate string and
string frame stacks.

There is an ANS Forth environmental dependence on lower case.

Either dstrings.fs or the pfe dstrings module must be loaded to
run this program.

For this program to work, these words need sufficient prior
testing:

  $DEPTH  $@  $! $S>
  $FRAME-DEPTH  MAX-#$FRAMES  $FSP0@  $FSP@  $FSP!

They have all been tested with earlier versions of dstester.fs and
dstrings-test.fs,
)

\ *** USAGE
(
PUBLIC INTERFACE WORDS

   tests:  ${  $->  }$
  config:  MAX-DST-INITIALS  MAX-DST-RESULTS  MAX-DSFT-RESULTS

PUBLIC AUXILIARY WORDS

  errors:  DST-ERROR-XT  DST-ERROR-DEFAULT
           DST-ERROR-INDEX  DSFT-ERROR-INDEX
  arrays:  DST-INITIALS  DST-RESULTS  DST-GOALS
           DST-A@  DST-A!  DST-AA=  DSFT-A!  DSFT-AA=

The public interface words are all that are needed for normal testing.

The public auxiliary words are made available for users who want to
implement more elaborate error reporting, such as counting errors, or
dumping results and goals when something is wrong, or reporting a
particular error with the help of DST-ERROR-INDEX or DSFT-ERROR-INDEX. 

The basic test usage takes the form, often on one line:
)
\    ( $initials) ${ <$results> <$fresults>
\            $-> <$goals> <$fgoals> }$ ( $initials)
(
Angle brackets around a data designation for the dynamic string or
dynamic string-frame stack indicate code that produces the data.  Here
$initials stands for possibly nonempty string stack contents at the
start of the test.  The initial string-frame stack must be empty.  The
sequence starting with ${ does the following:

  * Saves $initials and removes them from the $stack.

  * Executes the code <$results> and <$fresults> that produces the
    string stack $results and string-frame stack $fresults.

  * Saves and removes $results and $fresults from the stacks.

  * Executes the <$goals> and <$fgoals> code that produces the string
    stack $goals and string-frame stack $fgoals.

  * Saves and removes them from the stacks.

  * Compares the two sets of results and goals.

  * Reports when there is a discrepancy between the them.

  * Restores the saved $initials.

The words ${, $->, and }$ ignore all stacks except for the string and
string-frame stacks.

Many examples can be found in the files dstrings-test.fs and
dstester-test.fs.

Storage of initials, results, and goals for the two stacks includes
overflow checking.  The user can change the default storage allocation
by defining one or all of the following CONSTANT's *before* loading this
file:

  MAX-DST-INITIALS  MAX-DST-RESULTS  MAX-DST-RESULTS

The default execution token in the variable DST-ERROR-XT can be
changed only *after* loading this file.

Loading dstester.fs does not change BASE.
)
BASE @ DECIMAL

\ *** USER CONFIG

\ Defaults
0 [IF]
VARIABLE MAX-DST-INITIALS           32 MAX-DST-INITIALS !
VARIABLE MAX-DST-RESULTS            64 MAX-DST-RESULTS !
VARIABLE MAX-DSFT-RESULTS max-#$frames MAX-DSFT-RESULTS !
[THEN]
[UNDEFINED] MAX-DST-INITIALS [IF] 32 CONSTANT MAX-DST-INITIALS [THEN]
[UNDEFINED] MAX-DST-RESULTS  [IF] 64 CONSTANT MAX-DST-RESULTS [THEN]
[UNDEFINED] MAX-DSFT-RESULTS [IF]
                         max-#$frames CONSTANT MAX-DSFT-RESULTS [THEN]

\ *** ERRORS

VARIABLE DST-ERROR-INDEX   \ set to zero after unequal array size
VARIABLE DSFT-ERROR-INDEX  \ comparison, or [index+1] after deepest
                           \ unequal element comparison at index
VARIABLE DST-ERROR-XT
: DST-ERROR DST-ERROR-XT @ EXECUTE ;

: DST-ERROR-DEFAULT  ( c-addr u -- )  type source type cr ;
' DST-ERROR-DEFAULT DST-ERROR-XT !

\ *** ARRAYS
(
The first array slot has index zero.  A tick prefix means "address".
)

\ Array buffers begin with a cell-sized count.
CREATE DST-INITIALS MAX-DST-INITIALS 1+ cells allot
CREATE DST-RESULTS  MAX-DST-RESULTS  1+ cells allot
CREATE DST-GOALS    MAX-DST-RESULTS  1+ cells allot

\ $frame items are 2 cells containing a pointer into the $stack and
\ the size of the frame.
CREATE DSFT-RESULTS MAX-DSFT-RESULTS 2* 1+ cells allot
CREATE DSFT-GOALS   MAX-DSFT-RESULTS 2* 1+ cells allot

: 0DST-A  ( +n 'buf -- )  \ assume n<>0
(
Initialize the $array buffer for garbage collection by clearing the
count and n, $element addresses.
)
  swap 1+ cells over + swap DO 0 i ! 1 cells +LOOP ;

MAX-DST-INITIALS DST-INITIALS  0DST-A
MAX-DST-RESULTS  DST-RESULTS   0DST-A
MAX-DST-RESULTS  DST-GOALS     0DST-A

: $=  ($: a$ b$ -- ) ( -- flag )  $s> $s> compare 0= ;  \ wff
: empty-$fs  ( -- )  $fsp0@ $fsp! ;

\ Overflow checking to be done elsewhere.
: DST-A!   ( +n 'buf -- ) ( $: $[n-1] ... $0 -- )
  2dup !  cell+ swap  ( 'A_0 n) dup
  IF   cells over + swap DO i $! 1 cells +LOOP 
  ELSE 2drop THEN ;

: DST-A@   ( 'buf -- n ) ( $: -- $[n-1] ... $0 )
  dup @ dup ( n) >r ( 'buf n)
  IF   cell+ r@ 1- cells over + DO I $@ -1 cells +LOOP
  ELSE drop THEN r> ;

[UNDEFINED] CELL- [IF]
: CELL-  ( addr -- addr-cell )  [ 1 CELLS ] LITERAL - ;
[THEN]

: DST-AA=  ( 'buf1 'buf2 -- 0 true | 0 false | index+1 false )
(
Returns:
         0 true:  equal size and elements
        0 false:  sizes unequal
  index+1 false:  sizes equal, deepest unequal elements at index
)
  2dup @ swap @ <>
  IF 2drop 0 false EXIT THEN          \ unequal size
  dup @ 0= IF 2drop 0 true EXIT THEN  \ equal size and empty
  2>r 0 true 2r>
  ( 0 true 'buf1 'buf2)
  dup @ ( count) 1- ( last) >r
  cell+ r@ cells + swap
  cell+ r@ cells + 0 r>
  ( 0 true 'elems2_last 'elems1_last 0 last)
  DO
    2dup $@ $@ $= 0=
    IF  \ unequal elements
      2drop ( 0 true) 2drop i 1+ false 0 0 LEAVE THEN
    cell- swap cell-  \ [i-1]th element addresses interchanged 
  -1 +LOOP
  ( [0 true | i+1 false] x1 x2) 2drop ;

[UNDEFINED] DCELLS [IF]
: DCELLS ( n  -- 2*n*cell )  [ 2 CELLS ] LITERAL * ;
[THEN]
[UNDEFINED] DCELL+ [IF]
: DCELL+ ( x  -- x+[2*cell] )  [ 2 CELLS ] LITERAL + ;
[THEN]
[UNDEFINED] DCELL- [IF]
: DCELL- ( x  -- x-[2*cell] )  [ 2 CELLS ] LITERAL - ;
[THEN]

: DSFT-A!  ( +n 'buf -- ) ( $f: $f_[n-1] ... $f_0 --  )
  2dup !
  ( n 'buf) cell+ $fsp@
  LOCALS| source target |
  ( n)
  BEGIN dup
  WHILE
    source dup dcell+ to source 2@
    target dup dcell+ to target 2!
    1-
  REPEAT drop empty-$fs ;

: DSFT-AA=  ( 'buf1 'buf2 -- 0 true | 0 false | index+1 false )
(
Returns:
         0 true:  equal size and elements
        0 false:  sizes unequal
  index+1 false:  sizes equal, deepest unequal elements at index
)
  over @ over @ <>
  IF 2drop 0 false EXIT THEN          \ unequal size
  dup @ 0= IF 2drop 0 true EXIT THEN  \ equal size and empty
  2>r 0 true 2r>
  ( 0 true 'buf1 'buf2)
  dup @ ( count) 1- ( last) >r
  cell+ r@ dcells + swap
  cell+ r@ dcells + 0 r>
  ( 0 true 'A2_last 'A1_last 0 last)
  DO
    dup 2@ 2>r over 2@ 2r> d= 0=
    IF  \ unequal elements
      2drop ( 0 true) 2drop i 1+ false 0 0 LEAVE THEN
    dcell- swap dcell-  \ [i-1]th element addresses interchanged 
  -1 +LOOP
  ( [0 true | i+1 false] x1 x2) 2drop ;

: ${  ( $: $initials -- )
(
Throw an error if the string frame stack is not empty.  Save the
$initials for restoration by }$, leaving the string stack empty.  Clear
dstring results and goals storage, as well as storage for string frame
results and goals.
)
  $frame-depth ABORT" INITIAL $FRAME STACK NOT EMPTY"
  $depth dup MAX-DST-INITIALS > ABORT" TOO MANY $INITIALS"
  DST-INITIALS DST-A!   0 DST-RESULTS !    0 DST-GOALS !
  0 DSFT-RESULTS !   0 DSFT-GOALS ! ;

: $->  ( $: $results -- ) ( $f: $fresults -- )
(
Save the $results and $fresults, leaving the string and string frame
stacks empty.
)
  $depth dup MAX-DST-RESULTS > ABORT" TOO MANY $RESULTS"
  DST-RESULTS DST-A!
  $frame-depth dup MAX-DSFT-RESULTS > ABORT" TOO MANY $FRESULTS"
  DSFT-RESULTS DSFT-A! ; 

: }$  ( $: $goals -- $initials ) ( $f: $fgoals -- ) 
(
Save the $goals and $fgoals, compare them with the saved $results and
$fresults, and restore the $initials.  If the number of $results and/or
$fresults is not the same as the number of $goals and/or $fgoals, store
zero in DST-ERROR-INDEX and/or DSFT-ERROR-INDEX, respectively.  Else, if
the numbers are the same but there is a deepest string or string frame
comparison failure at the respective index-th element, store index+1 in
DST-ERROR-INDEX and/or DSFT-ERROR-INDEX.  Execute the BT-ERROR vector
once if any number or element comparison fails.
)
  $depth dup MAX-DST-RESULTS > ABORT" TOO MANY $GOALS"
  DST-GOALS DST-A!
  $frame-depth dup MAX-DSFT-RESULTS > ABORT" TOO MANY $FGOALS"
  DSFT-GOALS DSFT-A!

  DST-RESULTS DST-GOALS DST-AA= ( 0 true | 0 false | index+1 false)
  dup ( $ok?) >r
  IF
    ( 0) drop
  ELSE
    dup DST-ERROR-INDEX !
    IF S" INCORRECT $RESULT: " ELSE S" WRONG NUMBER OF $RESULTS: " THEN
  THEN  

  DSFT-RESULTS DSFT-GOALS DSFT-AA= ( 0 true | 0 false | index+1 false)
  dup ( $fok?) >r
  IF
    ( 0) drop
  ELSE
    dup DSFT-ERROR-INDEX !
    IF S" INCORRECT $FRESULT: " ELSE S" WRONG NUMBER OF $FRESULTS: " THEN
  THEN  

  2r> and 0= IF DST-ERROR THEN
  DST-INITIALS DST-A@ ( len) drop ;

BASE !
