(    Title:  Dynamic-Strings Word Set Tests
      File:  dstrings-test.fs
  Log file:  dstrings.log
    Author:  David N. Williams
   Version:  1.0.0
   License:  Public Domain
   Revised:  February 11, 2018

When used with dstrings.fs, this code is compatible with ANS Forth, with
an environmental dependence on lower case.  It can also be used to test
the pfe dstrings module.

This program calls dstester.fs, for which the following words need
sufficient prior testing:

  $DEPTH  $@  $! $S>
  $FRAME-DEPTH  MAX-#$FRAMES  $FSP0@  $FSP@  $FSP!

They have all been tested with earlier versions of dstester.fs and
dstrings-test.fs,

The code here often uses under-the-hood constructions not recommended
for normal code, like manipulating msa's [measured string addresses] on
the data stack, and looking at bla's [back link addresses].  It also
tests a few words which are not in the Dynamic-Strings word set, but
which are in both the pfe and ANS Forth implementations with the same
[or later] version number as this file.  Other under-the-hood words are
tested only implicitly.  The tests aim to be a good workout, but of
course do not prove the absence of bugs.

We use T{ ... -> ... }T instead of ={ ... => ... }= when we don't want a
test to empty the string stack, and we're not testing string stack
results.

Note that => records both data and string stacks, and }= compares both
data and string contents [not string addresses].

The independent program trydstrings.fs can also be used for
trouble-shooting.
)
decimal

\ debugging
[UNDEFINED] \\ [IF]
: \\  ( -- )  -1 parse 2drop BEGIN refill 0= UNTIL ;
[THEN]

s" ttester-xf.fs"     INCLUDED
s" tester-display.fs" INCLUDED  \ uses VERBOSE
s" xtester-errors.fs" INCLUDED
true VERBOSE !

\ Set the following to true to load DSTRINGS-EXT when pfe is the host.
false VALUE USE-DSTRINGS-EXT

PFE-HOST 0= [IF]
  false to USE-DSTRINGS-EXT
[THEN]

USE-DSTRINGS-EXT
[IF]   LOADM dstrings            \ pfe only
       : CHAR/  ( c[x] -- x )  1 chars / ;
[ELSE] s" dstrings.fs" INCLUDED  \ ANS Forth
[THEN]

4 CONSTANT max-#test-$frames
max-#test-$frames VALUE MAX-DSFT-RESULTS

s" dstester.fs" INCLUDED

s" ADDRESS-UNIT-BITS" environment? drop cells constant bits/cell
bits/cell 32 <>   bits/cell 64 <> and
[IF] cr .( ***Tests don't work unless bits/cell = 32 or 64.) ABORT [THEN]

bits/cell 64 = [IF] 26 cells [ELSE] 40 cells [THEN] constant /test-$buffer

: ={  T{ ${ ;   : =>  -> $-> ;   : }=  }T }$ ;

VARIABLE #errors    0 #errors !
: ?.errors  ( -- )
  VERBOSE @
  IF blue-text ." #ERRORS: " normal-text #errors @ . THEN ;

:noname  ( c-addr u -- )
(
Display an error message followed by the line that had the error. 
DST-ERROR-DEFAULT empties neither the string nor the string frame stack.
)
  red-text 1 #errors +! DST-ERROR-DEFAULT normal-text ;
DST-ERROR-XT !

: ?dump  ( c-addr u -- )
  VERBOSE @ IF dump [ PFE-HOST ] [IF] cr [THEN]
          ELSE 2drop THEN ;

\ before first displayed line
PFE-HOST IFORTH-HOST or [IF] ?.cr [THEN]

PFE-HOST [IF]
  ?." Host is pfe
[THEN]
GFORTH-HOST [IF]
  ?." Host is gforth
[THEN]
IFORTH-HOST [IF]
  ?." Host is iForth
[THEN]

USE-DSTRINGS-EXT [IF]
  ?." Testing with DSTRINGS-EXT module 
  : .tested  ( -- )  ." pfe DSTRINGS-EXT environment" ;
[ELSE]
  ?." Testing with dstrings.fs
  : .tested  ( -- )  ." dstrings.fs" ;
[THEN]

[UNDEFINED] cell- [IF]
  : cell-  ( addr -- addr-cell )  [ 1 cells ] literal - ; [THEN]
[UNDEFINED] 2cells [IF]
  2 cells constant 2cells [THEN]


\ *** TROUBLE-SHOOTING WORDS

: .$stat  ( -- )
  ." string buffer size: " /$buf . ." bytes"
      ."    max #$frames: " max-#$frames . cr
  ." $unused:  " $unused . ." bytes   "
      ." $depth: " $depth .
      ."    $frame depth: " $frame-depth . cr
  base @ hex
  ." $buffer:  0x" $bufp@ . cr
  ." $break:   0x" $breakp@ .
      4 spaces ." $frame-break:  0x" $fbreakp@ . cr
  ." $sp:      0x" $sp@ .
      4 spaces ." $frame-sp:     0x" $fsp@ . cr
  ." $sp0:     0x" $sp0@ .
      4 spaces ." $frame-sp0:    0x" $fsp0@ . cr
  $garbage? IF ." There is some string garbage. "
          ELSE ." There is no string garbage. " THEN cr
  cat$p@ IF ." Concatenation is in progress at: 0x" cat$p@ .
       ELSE ." No concatenation is in progress." THEN cr
  base ! ;

: ?dump-strings  ( -- )
  VERBOSE @ IF
    dstrings @ ( addr)
    /$buf /$space-header + /$frame-stack + cell+ ( #bytes)
    ?dump .$stat THEN ;

: ?dump-mstring  ( a$ -- )
  dup mcount@ /MCOUNT + aligned ?dump ;

: .stacks  .s ." $depth: " $depth . ;


\ *** TESTS

TESTING  /MCOUNT (M!) PARSE>S S` M,S

s" This is a measured string with 45 characters."
    m,s 2constant m1-s

?." Measured string count-field size: /MCOUNT = "
  /MCOUNT ?. ?.cr
?." You should see a string duplicated with the same
?." alignment and between the same quotes:
?." This is a measured string with 45 characters."
?.( ") m1-s ?type ?.( ") ?.cr

?." Check that the measured string dump has a 0x2D count and the
?." right count-field size, and is zero filled to alignment: "
m1-s -mcount ?dump-mstring
?." The address just after the zero padding should be: "
   m1-s + aligned hex ?. decimal ?.cr

?." The following dump should look the same as the above, except
?." for its address and possible trailing x's: "
pad 64 char x fill
T{ m1-s pad (m!) -> }T
pad ?dump-mstring

s` This is a measured string with 45 characters.`
    m,s 2constant m2-s  \ M,S because could be transient

: m3-s  ( -- s  )
  s` This is a measured string with 45 characters.` ;

: s^  ( "ccc<^>" char -- )  [char] ^ parse>s ; immediate

char ^ parse>s This is a measured string with 45 characters.^
    m,s 2constant m4-s

: m5-s  ( -- s )
  s^ This is a measured string with 45 characters.^ ;

T{ m2-s m1-s compare -> false }T
T{ m3-s m1-s compare -> false }T
T{ m4-s m1-s compare -> false }T
T{ m5-s m1-s compare -> false }T

TESTING  DSTRINGS MAKE-$SPACE /$FRAME-ITEM MAX-#$FRAMES /$FRAME-STACK
(
Replace the default string space with a smaller one for testing.
)
  dstrings @ free ( ior) THROW
  /test-$buffer max-#test-$frames make-$space dstrings !

={ max-#$frames  => max-#test-$frames }=
T{ /$frame-stack -> max-#$frames /$frame-item * }T 

TESTING  0STRINGS with string space empty

0strings
?." Empty string space used for testing:"
?dump-strings

TESTING  MCOUNT -MCOUNT $. $TYPE >$S $S> $POP $PUSH-EXT

={ m1-s >$s => m1-s -mcount $push-ext }=
T{ m1-s -mcount mcount -> m1-s }T
={ m1-s >$s $s>  => m1-s }=
={ m1-s >$s $pop => m1-s -mcount }=
={ m1-s -mcount $push-ext => m1-s >$s }=
={ m1-s >$s $pop mcount   => m1-s }=

: m$  ($: -- a$ )  m1-s >$s ;

?." You should see the same string on the next two lines without quotes:
VERBOSE @ [IF]
={ m$ $. => }= cr
={ m$ $type => }= cr
[THEN]

TESTING  $" $` $CONSTANT $VARIABLE MCOUNT@ MCOUNT! PARSE>$

?." You should see a string three times with the same
?." alignment and between the same quotes:
VERBOSE @ [IF]
.( "This is a short string.") cr
.( ") ={ $" This is a short string." $. => }= .( ") cr
: ."short"  ( -- ) $" This is a short string." $. ;
.( ") ={ ."short" => }= .( ") cr
[THEN]

$" Another short string." $constant ashort$1

: ashort$2  ($: -- a$ )  $" Another short string." ;

$` Another short string.` $constant ashort$3
: ashort$4  ($: -- a$ )  $` Another short string.` ;

char ^ parse>$ Another short string.^ $constant ashort$5

: $^  ( "ccc<^>" -- $: a$ )  [char] ^ parse>$ ; immediate
: ashort$6  ($: -- a$)  $^ Another short string.^ ;

={ ashort$1 => $" Another short string." }=
={ ashort$1 => ashort$2 }=
={ ashort$1 => ashort$3 }=
={ ashort$1 => ashort$4 }=
={ ashort$1 => ashort$5 }=
={ ashort$1 => ashort$6 }=

={ : my$con  $constant ; => }=
={ ashort$1 my$con my-ashort$ => }=
={ ashort$1 => my-ashort$ }=

={ $variable var$ => }=
={ var$ @ mcount@ => 0 }=  \ empty$

ashort$1 $pop var$ !

\ The following line changes an external string, often a no-no:
={ 4 var$ @ mcount! var$ @ $push-ext => $" Anot" }=

={ : my$var  $variable ; => }=
={ my$var avar$  avar$ @ $push-ext => empty$ }=

TESTING  \N$ EMPTY$

={ empty$ $pop mcount@ => 0 }=
={ \n$ $s> swap c@ => 1 10 }=

TESTING  $@ $! >$S >$S-COPY $S> $,S $S@ IN-$BUFFER?

\ Note that $! is already used in dstester.fs, but up to now
\ only with external strings.

={ empty$ var$ $! => }=
={ var$ $@ => empty$ }=
ashort$1 var$ $!
={ var$ $@ => ashort$1 }=

s" short" m,s 2constant short-s
$" short"     $constant short$ 

={ short-s >$s => short-s -mcount $push-ext }=
={ short-s >$s-copy var$ $! => }=
T{ var$ @ in-$buffer? -> true }T
T{ short-s -mcount in-$buffer? -> false }T

\ The following should produce a second copy of "short" in the
\ string buffer, because => does a $!, and var$ already has a
\ copy in the string buffer:
={ var$ $@ => short$ }=

={ short$ $s> => short$ $pop mcount }=
={ short$ $s@ => short$ $pop mcount short$ }=
$" hello" $constant hello$
={ here hello$ $,s 2constant hello-s => hello-s -mcount }=

?." You should see a correct dump of 'hello' as a measured string,
?." zero-filled to alignment:"
hello-s -mcount ?dump-mstring

TESTING  0STRINGS 0$SPACE

\ Test 0STRINGS with garbage and a string bound to VAR$:

s" garbage" >$s-copy $drop
s" bound"   >$s-copy var$ $!
T{ var$ @ in-$buffer? $garbage? -> true true }T
={ short$ 0strings $sp@ $bufp@ var$ $@ => $sp0@ $breakp@ empty$ }=
={ $unused $depth $frame-depth => /test-$buffer 0 0 }=

\ Test 0$SPACE with garbage and a string bound to VAR$:

0strings
variable tscratch
s" garbage" >$s-copy $drop
s" bound"   >$s-copy var$ $!
short$                       \ another string on the $stack
var$ @ tscratch !
T{ var$ @ in-$buffer? $garbage? -> true true }T
T{ dstrings @ 0$space $sp@ $bufp@ $depth -> $sp0@ $breakp@ 0 }T
T{ var$ @ -> tscratch @ }T  \ 0$SPACE leaves VAR$ dangling
empty$ $pop var$ !           \ undangle it
={ $unused $frame-depth => /test-$buffer 0 }=

TESTING  $DROP $2DROP $DUP $2DUP $SWAP $NIP $OVER $TUCK $PICK $EXCHANGE

hello$ $constant h$   short$ $constant s$
0strings

={ h$       $drop  =>    }=
={ h$ h$    $drop  => h$ }=
={ h$ h$    $2drop =>    }=
={ h$ h$ h$ $2drop => h$ }=

={ h$ $dup     => h$ h$ }=
={ h$ s$ $2dup => h$ s$ h$ s$ }=

={ h$ s$ $swap => s$ h$ }=
={ h$ s$ $nip  => s$ }=
={ h$ s$ $over => h$ s$ h$ }=
={ h$ s$ $tuck => s$ h$ s$ }=

={ h$    0 $pick => h$ h$ }=
={ h$ s$ 1 $pick => h$ s$ h$ }=


$" a" $constant a$
$" b" $constant b$

={ h$ s$ 0 1 $exchange => s$ h$ }=
={ h$ s$ 1 0 $exchange => s$ h$ }=
={ h$ a$ s$ s$    1 3 $exchange => s$ a$ h$ s$ }=
={ h$ a$ b$ s$ a$ 4 1 $exchange => s$ a$ b$ h$ a$ }=

TESTING  bound string back-link and auto-copy rule
(
The auto-copy rule is that the only time a bound string is
copied without explicit demand is by $! when it stores a string
already bound to a string variable into another string variable.
)

0strings
\ These two lines should leave a copy of "hello" bound to VAR$
\ and a copy of "short" bound to the string stack:

  h$ $s> >$s-copy var$ $!
  var$ $@ s$ $s> >$s-copy  ($: h$ s$)

\ binding tests

: blink@  ( msa -- blink )  cell- @ ;
T{ var$ @ in-$buffer?         -> true }T
T{ var$ @ blink@              -> var$ }T
T{ $sp@ @ blink@              -> $sp@ }T
T{ var$ $@ $s@ -mcount blink@ -> var$ }T   ($: h$ s$ h$)
T{ s$ var$ $! $depth          -> 3 }T      ($: h$ s$ h$)
T{ $sp@ @ blink@              -> $sp@ 2cells + }T
T{ $2drop $sp@ @ blink@       -> $sp@ }T   ($: h$)
T{ $garbage?                  -> true }T
T{ $sp@ cell- @ blink@        -> 0 }T      \ s$ copy is garbage
T{ var$ $@ $sp@ @ in-$buffer? -> false }T  ($: h$ s$-ext)
$drop                                      ($: h$)

\ copying tests

\ bound h$ not copied when stored on top of external s$ 
T{ $sp@ @ var$ $! $depth  -> var$ @ 0 }T
\ nor when stored on top of itself
={ var$ @ var$ $@ var$ $! => var$ @ }=

$variable var2$   var$ $@ var2$ $!   \ should make a new copy of h$
T{ var$ @  in-$buffer? -> true }T    \ old copy in string buffer
T{ var2$ @ in-$buffer? -> true }T    \ new copy in string buffer  
T{ var$ @  var2$ @ <>  -> true }T    \ new copy distinct from old
={ var$ $@ => var2$ $@ }=            \ but really is a copy
\ note that => may have generated one or more copies as well

\ USE-DSTRINGS-EXT 0= [IF]
TESTING  nonexternal $CONSTANT copying

\ nonext is copied
m1-s >$s-copy
T{ $garbage? -> false }T
T{ $s@ -mcount in-$buffer? $depth -> true 1 }T
T{ $constant m4$ $depth -> 0 }T
T{ $garbage? $sp@ cell- @ blink@ -> true 0 }T
={ m4$ $s> -mcount in-$buffer? => false }=
={ m4$ $s> -mcount m1-s -mcount <> => true }=
={ m4$ => m1-s >$s }=

\ ext is not copied
m1-s >$s
T{ $s@ -mcount in-$buffer? $depth -> false 1 }T
T{ $constant m5$ $depth -> 0 }T
={ m5$ $s> -mcount in-$buffer? => false }=
={ m5$ $s> -mcount m1-s -mcount = => true }=
={ m5$ => m1-s >$s }=
\ [THEN]

TESTING  $SWAP and $EXCHANGE relinking

\ We don't claim to have exercised all paths in the code.

0strings

a$ $s> >$s-copy var$ $!

\ strings bound to string stack
: a-bnd$  ($: -- a$ )  s" a" >$s-copy ;
: b-bnd$  ($: -- b$ )  s" b" >$s-copy ;

: th-blink  ( i $: a_i$ ... -- a_i$ ... s: blink_i )
  $pick $s> -mcount blink@ ;

={ var$ $@ s$ $swap var$ @ blink@ => s$ var$ $@ var$ @ blink@ }=
={ s$ var$ $@ $swap var$ @ blink@ => var$ $@ s$ var$ @ blink@ }=

={ a$ b-bnd$ 0 th-blink $swap 1 th-blink
    => b$ a$ $sp@ dup cell+ }=

={ a-bnd$ b-bnd$ 0 th-blink 1 th-blink $swap
    0 th-blink 1 th-blink
      => b$ a$ $sp@ dup cell+ 2dup }=

={ a-bnd$ b-bnd$ $over 1 th-blink 0 th-blink $swap
    0 th-blink 1 th-blink
      => a$ a$ b$ $sp@ cell+ dup cell+ $sp@ dup 2 cells + }=

={ a$ b-bnd$ 0 th-blink 0 1 $exchange 1 th-blink
      => b$ a$ $sp@ dup cell+ }=

={ a-bnd$ b-bnd$ 0 th-blink 1 th-blink 1 0 $exchange
    0 th-blink 1 th-blink
      => b$ a$ $sp@ dup cell+ 2dup }=

={ b-bnd$ b-bnd$ a$ 1 th-blink 2 th-blink 1 2 $exchange
    1 th-blink 2 th-blink
      => b$ b$ a$ $sp@ cell+ dup cell+ 2dup }=

={ a-bnd$ h$ s$ b-bnd$ h$ 1 th-blink 4 th-blink 1 4 $exchange
  1 th-blink 4 th-blink
    => b$ h$ s$ a$ h$ $sp@ cell+ dup 3 cells + 2dup }=

TESTING  $+ S+ PARSE-S+ $+" $+` ENDCAT

0strings
: bl+           ( -- )  $+"  " ;
: $+`this_is_`  ( -- )  $+` this is ` ;
: $+^  ( -- )   ( -- )  [char] ^ parse-s+ ; immediate
: $+^this_is_^  ( -- )  $+^ this is ^ ;
: cat?  ( -- flag )  cat$p@ 0= invert ;

$" short hello"   $constant s_h$
$" this is short" $constant d$

={ s$ $+ bl+ h$ $+    endcat => s_h$ }= 
={ $+`this_is_` s$ $+ endcat => d$ }=
={ $+^this_is_^ s$ $+ endcat => d$ }=

={ s$ $+ $+"  " $+" hello" endcat => s_h$ }=
={ $+" short" bl+ h$ $+ endcat    => s_h$ }=
={ s$ $+ $+`  ` $+` hello` endcat => s_h$ }=
={ $+` short` bl+ h$ $+ endcat    => s_h$ }=

T{ cat? empty$ $+ cat? -> 0 0 }T
T{ empty$ $+ short-s >$s-copy $drop -> }T
T{ cat? s" " s+ cat? -> 0 0 }T
T{ cat? $+" " cat? -> 0 0 }T
T{ cat? $+` ` cat? -> 0 0 }T
T{ cat? :noname $+` `; execute cat? -> 0 0 }T

0strings
={ s" this is " s+ s" short" s+ endcat => d$ }=
={ char ^ dup parse-s+ this is ^ parse-s+ short^ endcat => d$ }=
={ s$ $+ => }=
={ bl+ h$ $+ => }=
={ endcat => s_h$ }=
={ endcat => empty$ }=
T{ cat? char ^ parse-s+ ^ cat? -> 0 0 }T
T{ cat? :noname [char] ^ parse-s+ ; execute ^ cat? -> 0 0 }T

0strings
={ short-s >$s-copy $dup $+ bl+ h$ $+ ENDCAT => s$ s_h$ }=

\ Test that ENDCAT turns off the cat lock and that the cat$
\ gets marked as garbage by $DROP (probably superfluous):

0strings
T{ h$ $+ cat$p@ endcat cat$p@ $drop swap blink@ -> 0 0 }T

TESTING  $FRAME DROP-$FRAME FIND-$ARG $FRAME-DEPTH

0strings

={ $frame-depth => 0 }=
T{ 0 $frame $depth $frame-depth -> 0 1 }T
T{ h$ s$ s_h$ 3 $frame $frame-depth -> 2 }T
T{ h$   $s> find-$arg -> 2 true }T
T{ s$   $s> find-$arg -> 1 true }T
T{ s_h$ $s> find-$arg -> 0 true }T

T{ d$ $s> find-$arg drop-$frame $depth $frame-depth -> false 0 1 }T

\ Repeat with the string frame below the top string:
T{ h$ s$ s_h$ 3 $frame $frame-depth d$ -> 2 }T
T{ h$   $s> find-$arg -> 2 true }T
T{ s$   $s> find-$arg -> 1 true }T
T{ s_h$ $s> find-$arg -> 0 true }T

T{ $s> find-$arg drop-$frame $frame-depth $depth -> false 1 0 }T
T{ drop-$frame $frame-depth -> 0 }T  \ dropped frame had zero items

0strings
T{ h$ s$ s_h$ 3 $frame $frame-depth -> 1 }T
0strings
T{ h$ s$ s_h$ 3 $frame d$ 1 $frame $frame-depth -> 2 }T
0strings
T{ $frame-depth $fsp@ $sp@ $breakp@ -> 0 $fsp0@ $sp0@ $bufp@ }T

\ Drop string frame below other strings:
0strings
={ 0 $frame a$ drop-$frame $frame-depth => a$ 0 }=
={ a$ b$ 2 $frame s$ drop-$frame $frame-depth => s$ 0 }=
={ a-bnd$ b-bnd$ 2 $frame s$ drop-$frame $frame-depth => s$ 0 }=
={ a$ 1 $frame b$ s$ drop-$frame $frame-depth => b$ s$ 0 }=
={ a-bnd$ 1 $frame b-bnd$ s$ drop-$frame $frame-depth => b$ s$ 0 }=
={ a$ b$ 2 $frame s$ 1 $frame h$ drop-$frame
     => a$ b$ 2 $frame h$ }=
0strings
={ h$ s$ 2 $frame a-bnd$ b-bnd$ 2 $frame h$ drop-$frame
   collect-$garbage
     => h$ s$ 2 $frame h$ true }=

TESTING  COLLECT-$GARBAGE $GC-OFF $GC-ON $GC-LOCK@ $GC-LOCK!

0strings
s" goodbye" m,s 2constant goodbye-s
s" junk"    m,s 2constant junk-s

: $space-clear?  ( -- flag )
  $frame-depth 0=   $fsp@ $fsp0@ = and   $sp@ $sp0@ = and
      $breakp@ $bufp@ = and   cat$p@ 0= and ;

(
Make a garbage hole at the beginning of the string buffer, a
hole between bound strings, and a hole as the last string in the
buffer, with some duplication on the string stack, and something
stored in a string variable.
)
  short$
  junk-s    >$s-copy $drop         \ beginning hole
  hello-s   >$s-copy $dup
  junk-s    >$s-copy $drop         \ double middle hole
  junk-s    >$s-copy $drop
  goodbye-s >$s-copy $dup var$ $!
  junk-s    >$s-copy $drop         \ end hole

T{ $garbage? -> true }T 

?." You should see a garbage hole at the start of the string buffer,
?." a string bound to and duplicated on the string stack, a garbage
?." hole with two strings, a string bound to a variable and
?." duplicated on the stack, and a garbage hole with one string.
?." The fourth, deepest string on the stack is external."
?dump-strings

s" This is a new string." >$s-copy

?." Now we have triggered a garbage collection by copying in
?." another string, so you should see the holes removed."
?dump-strings

T{ collect-$garbage -> false }T
T{ $2drop $2drop $drop $depth -> 0 }T

\ Now it's all garbage except for the next to last string, which
\ is stored in var$.
={ $garbage? collect-$garbage $garbage? => true true false }=

\ Only the var$ is left.
T{ var$ @ in-$buffer? -> true }T
={ empty$ var$ $! $garbage? => true }=

\ Now it's garbage, too.
={ collect-$garbage $garbage? => true false }=
={ $space-clear? => true }=

(
Next we test COLLECT-$GARBAGE with the first string alive,
followed by several garbage strings.
)
  0strings
  hello-s   >$s-copy
  junk-s    >$s-copy $drop
  goodbye-s >$s-copy $drop
  junk-s    >$s-copy $drop

T{ $depth collect-$garbage $drop collect-$garbage $depth -> 1 true true 0 }T
T{ $space-clear? -> true }T

(
Test with nothing but garbage.
)
  0strings
  hello-s   >$s-copy $drop
  junk-s    >$s-copy $drop
  goodbye-s >$s-copy $drop

T{ collect-$garbage $space-clear? -> true true }T

\ GC LOCK TEST

\ Make some garbage.
={ hello-s >$s-copy $drop $gc-off => }=
T{ $gc-lock@ -> true }T

={ $gc-on collect-$garbage => true }=
T{ $gc-lock@ -> false }T

T{ true  $gc-lock! $gc-lock@ -> true }T
T{ false $gc-lock! $gc-lock@ -> false $gc-on }T

TESTING  #$ARGS $ARGS{ TH-$ARG

0strings

T{ h$ s$ s_h$ 3 $frame #$args -> 3 }T

T{ s_h$ 0 th-$arg $s> $s> compare -> 0 }T
T{ h$   2 th-$arg $s> $s> compare -> 0 }T 
T{ s$ #$args $frame-depth $depth  -> 3 1 4 }T

\ Example from documentation:
: george  ($: a$ b$ c$ -- cat$ )
      $ARGS{ arg1 arg2 arg3 }
      $+" This is arg1:  " arg1 $+" ." ENDCAT ;

0strings
={ $" bill"  $" sue"  $" marie"  george =>
    $" This is arg1:  bill." }=

: 3args  ($: a$ b$ c$ -- )
  $+" Arguments: " $ARGS{ 
                           alpha beta
                           gamma
                         }
  $+` one = "` alpha $+` ", two = "` beta
  $+` " three = "` gamma $+` ".` ENDCAT
  alpha beta gamma ENDCAT avar$ $! var$ $! ;

0strings
$" What " $" a " $" blast!" 3args
={ $` Arguments: one = "What ", two = "a " three = "blast!".`
    => var$ $@ }=

TESTING  exceptions other than string buffer overflow
(
Other exception not tested: "dynamic string count too large". It
could be removed.

The tests in this and the buffer overflow section implicitly,
albeit very weakly, verify the rule that exceptions leave string
space in a valid state.  Often they leave string space
unchanged, including the stacks, from its state just before
execution of the word, up to garbage collection.  But the
dstrings.fs implementation especially is allowed to temporarily
push onto the string or string frame stack, so something extra
may be there at the point of an exception.

There is no guarantee for the data stack, but the CATCHes in the
tests take care of that by restoring the depth from the
exception frames they push.

Tested:  $POP $S> $DROP $2DROP $DUP $2DUP $SWAP
         $NIP $OVER $TUCK $PICK $EXCHANGE
         $FRAME DROP-$FRAME TH-$ARG $ARGS{ #$ARGS
         >$S-COPY COLLECT-$GARBAGE
)

USE-DSTRINGS-EXT [IF]
  -2053 CONSTANT SCOUNT-OV
  -2054 CONSTANT $BUF-OV
  -2055 CONSTANT $GARBAGE-LOCK
  -2056 CONSTANT $S-UN
  -2057 CONSTANT $CAT-LOCK
  -2058 CONSTANT DSCOUNT-OV
  -2059 CONSTANT $FS-OV
  -2060 CONSTANT $F-IOV
  -2061 CONSTANT $FS-UN
[ELSE]
  -2 CONSTANT SCOUNT-OV 
  -2 CONSTANT $BUF-OV
  -2 CONSTANT $GARBAGE-LOCK
  -2 CONSTANT $S-UN
  -2 CONSTANT $CAT-LOCK
  -2 CONSTANT DSCOUNT-OV
  -2 CONSTANT $FS-OV
  -2 CONSTANT $F-IOV
  -2 CONSTANT $FS-UN
[THEN]

(
Uncomment one of the lines below to see that the exception
message is correct.  The trailing comments in some of the lines
show pfe system messages, displayed when USE-DSTRINGS-EXT is
true.
)

: nip2  nip nip ;

/MCOUNT 1 cells < [IF]
\ 0 MAX-MCOUNT 1+ m,s  \ string count too large
T{ 0 MAX-MCOUNT 1+ ' m,s CATCH nip2 -> SCOUNT-OV }T
\ 0 MAX-MCOUNT 1+ >$s-copy
T{ 0 MAX-MCOUNT 1+ ' >$s-copy CATCH nip2 -> SCOUNT-OV }T
\ 0 MAX-MCOUNT 1+ s+
T{ 0 MAX-MCOUNT 1+ ' s+ CATCH nip2 -> SCOUNT-OV }T
[THEN]

\ $pop  \ string stack underflow
T{ ' $pop CATCH -> $S-UN }T
\ $s>
T{ ' $s> CATCH -> $S-UN }T
\ $drop
T{ ' $drop CATCH -> $S-UN }T
\ $2drop
T{ ' $2drop CATCH -> $S-UN }T
\ h$ $2drop
T{ h$ ' $2drop CATCH -> $S-UN 0strings }T
\ $dup
T{ ' $dup CATCH -> $S-UN }T
\ $2dup
T{ ' $2dup CATCH -> $S-UN }T
\ h$ $2dup
T{ h$ ' $2dup CATCH -> $S-UN 0strings }T
\ $swap
T{ ' $swap CATCH -> $S-UN }T
\ h$ $swap
T{ h$ ' $swap CATCH -> $S-UN 0strings }T
\ $nip
T{ ' $nip CATCH -> $S-UN }T
\ h$ $nip
T{ h$ ' $nip CATCH -> $S-UN 0strings }T
\ $over
T{ ' $over CATCH -> $S-UN }T
\ h$ $over
T{ h$ ' $over CATCH -> $S-UN 0strings }T
\ $tuck
T{ ' $tuck CATCH -> $S-UN }T
\ 0 $pick
T{ 0 ' $pick CATCH nip -> $S-UN }T
\ h$ 1 $pick
T{ h$ 1 ' $pick CATCH nip -> $S-UN 0strings }T
\ 0 0 $exchange
T{ 0 0 ' $exchange CATCH nip2 -> $S-UN }T
\ h$ 1 0 $exchange
T{ h$ 1 0 ' $exchange CATCH nip2 -> $S-UN 0strings }T
\ h$ $dup $dup 0 3 $exchange
T{ h$ $dup $dup 0 3 ' $exchange CATCH nip2 -> $S-UN 0strings }T
\ h$ $dup $dup 3 0 $exchange
T{ h$ $dup $dup 3 0 ' $exchange CATCH nip2 -> $S-UN 0strings }T
\ h$ $dup $dup 3 3 $exchange
T{ h$ $dup $dup 3 3 ' $exchange CATCH nip2 -> $S-UN 0strings }T

\ 0 $frame 0 $frame 0 $frame 0 $frame 0 $frame \ too many string frames
T{ 0 $frame 0 $frame 0 $frame 0 $frame
   0 ' $frame CATCH nip -> $FS-OV 0strings }T
\ drop-$frame  \ string frame stack underflow
T{ ' drop-$frame CATCH -> $FS-UN }T
\ 1 $frame \ string stack underflow
T{ 1 ' $frame CATCH nip -> $S-UN }T
\ 0 $frame h$ 2 $frame
T{ 0 $frame h$ 2 ' $frame CATCH nip -> $S-UN 0strings }T

\ $ARGS{  \ interpreting a compile-only word
T{ ' $ARGS{ CATCH -> -14 }T
\ #$args  \ string frame stack underflow
T{ ' #$args CATCH -> $FS-UN }T
\ 0 th-$arg
T{ 0 ' th-$arg CATCH nip -> $FS-UN }T
\ h$ $dup 2 $frame h$ 2 th-$arg  \ not enough strings in top frame
T{ h$ $dup 2 $frame h$ 2 ' th-$arg CATCH nip -> $F-IOV 0strings }T

\ h$ $+ s$ >$s-copy  \cat lock preventing string copy
T{ h$ $+ s$ ' >$s-copy CATCH -> $CAT-LOCK }T
={ ENDCAT $drop $gc-off => }=
\ $gc-off collect-$garbage \ string garbage locked
T{ ' collect-$garbage CATCH -> $GARBAGE-LOCK }T

TESTING  string buffer overflow exceptions
(
The following constructs a minimal space for testing string
space overflow.  The original test space stays around, and holds
a string used to saturate the minimal space, and another string
exactly one character longer.  As far as the minimal space is
concerned, these are external, constant strings.

Tested:  EMPTY$ $CONSTANT >$S-COPY
         $DUP $2DUP $OVER $PICK
         $@ $PUSH-EXT >$S $" $` PARSE>$ $!
         $+ S+ $+" $+` PARSE-S+ ENDCAT
         TH-$ARG
)

$VARIABLE minx$   $VARIABLE minx+y$

: x$  ( $: -- x$ )
(
Generate the minimum string of "x"'s whose body occupies at
least one cell, and which is aligned, including its count field.
)
  /mcount aligned /mcount - cell+ char/ 0
  ?DO $+" x" LOOP ENDCAT ;

0strings
x$ $dup minx$ $!   ($: x$) $+ $+" y" ENDCAT minx+y$ $!
$" " $CONSTANT 0$
$gc-off                       \ freeze strings in old space
dstrings @ value test-$space  \ keep old space

(
Make a string space just big enough to hold one copy of the
"x"'s string, two string stack items, and one string frame.
)
  x$ $s> nip /mcount + 3 cells + ( /$buf) 1 ( #$frames)
  make-$space dstrings !  \ new, minimal space

\ copy to string buffer without space overhead
: "x..x"   ( $: -- x..x$ )   minx$   @ mcount >$s-copy ;
: "x..xy"  ( $: -- x..xy$ )  minx+y$ @ mcount >$s-copy ;
: "z"      ( $: -- z$ )      s" z" >$s-copy ;

: fill$buf  ( $: i*$ -- x..x$ x..x$ )  0strings "x..x" $dup ;

fill$buf
?." You should see a small string space with a full string
?." buffer containing one bound string consisting of x's to
?." trailing alignment, referenced twice on the string stack."
?dump-strings

\ empty$  \ dstring buffer overflow
T{ ' empty$ CATCH -> $BUF-OV }T  \ sys $constant, $s ov
T{     ' 0$ CATCH -> $BUF-OV }T  \ $constant, $s ov

T{ 0strings 0$   ' "x..xy" CATCH -> $BUF-OV }T \ >$S-COPY, body ov
T{ fill$buf $drop ' "x..x" CATCH -> $BUF-OV }T \ >$S-COPY, $s ov

\ None of the words tested in the following copies a string. 
\ It's the string stack that overflows.
fill$buf
T{ ' $dup  CATCH -> $BUF-OV }T
T{ ' $2dup CATCH -> $BUF-OV }T
T{ ' $over CATCH -> $BUF-OV }T
T{  1 ' $pick CATCH nip -> $BUF-OV }T
T{ minx$ ' $@ CATCH nip -> $BUF-OV }T
T{ s" x" ' $push-ext CATCH nip2 -> $BUF-OV }T
T{ s" y"       ' >$s CATCH nip2 -> $BUF-OV }T
: $"z"  $" z" ;   : $`z`  $` z`;   \ ' $" CATCH z" doesn't work with pfe
: $^z^  $^ z^ ;   \ $^ defined above in terms of PARSE>$
T{ ' $"z" CATCH -> $BUF-OV }T
T{ ' $`z` CATCH -> $BUF-OV }T
T{ ' $^z^ CATCH -> $BUF-OV }T  \ tests PARSE>$

\ Force $! to copy:
T{ tscratch $! $dup var$ ' $! CATCH nip -> $BUF-OV }T

\ concatenation
T{ fill$buf $drop $" z" ' $+ CATCH -> $BUF-OV }T
T{ fill$buf $drop s" z" ' s+ CATCH nip2 -> $BUF-OV }T
: $+"z"  $+" x" ;   : $+`z`  $+` x` ;
: $+^z^  $+^ z^ ;   \ $+^ defined above in terms of PARSE-S+
T{ fill$buf $drop ' $+"z" CATCH -> $BUF-OV }T
T{ fill$buf $drop ' $+`z` CATCH -> $BUF-OV }T
T{ fill$buf $drop ' $+^z^ CATCH -> $BUF-OV }T  \ tests PARSE-S+
0strings minx$ @ mcount s+ 0$ 0$  \ $buf full with cat$
T{ ' ENDCAT CATCH -> $BUF-OV }T

T{ fill$buf 2 $frame 1 ' th-$arg CATCH nip -> $BUF-OV }T

dstrings @ free THROW  \ done with overflow test space
test-$space dstrings !   $gc-on

VERBOSE @ [IF] blue-text
avar$ $@ $.
.( Done testing ) .tested .(  with /MCOUNT = )
    /MCOUNT 1 u.r .( .) cr
.( #ERRORS: ) #errors @ .
GFORTH-HOST [IF] cr [THEN]
normal-text [THEN]
